#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>

// Define here the type of test
#define PHYSMEM_TEST
// #define IPC_TEST

typedef unsigned long long wordType;

struct timeval now, start, end;

volatile void *memPtr = NULL;
unsigned long long memSizeBytes = 0;
unsigned long long currSizeBytes = 0;

#ifdef IPC_TEST
int shmid = -1;
#endif

char *thisFile = NULL;

#define TEST_WARMUP        2
#define TEST_MIN_DURATION  3000000
#define TEST_MIN_LOOPS     3

#define MAX_TEST_SIZE_BYTES (long long)((long long)400.0 * 1024.0 * 1024.0)

#ifndef TRUE
#  define TRUE (0 == 0)
#endif
#ifndef FALSE
#  define FALSE (0 == 1)
#endif
#define SP(l) l,sizeof(l)
#define AP(l) &l[strlen(l)],sizeof(l)-strlen(l)

// ===========================================================
#ifdef PHYSMEM_TEST
int physmemFd = -1;
#endif

inline void openMem( unsigned long long mapSizeBytes ) {
  if ( memPtr != NULL )
    fprintf( stderr, "ERROR: openMem called with memory already mapped (ignored)\n" );

  currSizeBytes = mapSizeBytes;
#ifdef PHYSMEM_TEST
  if ( physmemFd != -1 )
    fprintf( stderr, "ERROR: openMem called with PHYSMEM already open (ignored)\n" );

  if ( (physmemFd = open( "/dev/physmem1", O_RDWR )) == -1 ) {
    perror( "Failed to open physmem" );
    exit( 1 );
  }
  int ioctlStat = 0;
#define PHYSMEM_GETSIZE  _IOR('x', 2, u_long)
  if( (ioctlStat = ioctl( physmemFd, PHYSMEM_GETSIZE, &memSizeBytes )) < 0  ) {
    perror( "Failed to get PHYSMEM size" );
    exit( 1 );
  }
  if ( mapSizeBytes == 0 ) return;
  if ( mapSizeBytes > memSizeBytes ) {
    fprintf( stderr, "PHYSMEM too small needed:%lld got:%lld\n", mapSizeBytes, memSizeBytes );
    exit( 1 );
  }
  // mmap( 0, 4, PROT_READ|PROT_WRITE, MAP_SHARED, physmemFd, 0 );
  // fprintf( stderr, "Mapping for %llu (%016llx) bytes\n", mapSizeBytes, mapSizeBytes ); fflush( stderr );
  if ( (memPtr = mmap( 0, mapSizeBytes, PROT_READ|PROT_WRITE, MAP_SHARED, physmemFd, 0 )) == MAP_FAILED ) {
    perror( "Failed to map physmem" );
    exit( 1 );
  }
  return;
#endif
#ifdef IPC_TEST
  static key_t fkey = -1;

  if ( fkey == -1 ) {
    if ( (fkey = ftok( thisFile, 1 )) == -1 ) {
      perror( "ftok for IPC failed" );
      exit( 1 );
    }
  }
  if ( mapSizeBytes == 0 ) {
    memSizeBytes = (unsigned long long)10 * 1024 * 1024 * 1024;
    return;
  }
  if ( (shmid = shmget( fkey, mapSizeBytes, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH |IPC_CREAT )) == -1 ) {
    perror( "Failed to create IPC block" );
    fprintf( stderr, "IPC size:%llu\n", memSizeBytes );
    exit( 1 );
  }
  if ( (memPtr = shmat( shmid, NULL, 0 )) == (void *)-1 ) {
    perror( "Failed to attach to IPC block" );
    fprintf( stderr, "errno:%d (%s)", errno, strerror( errno ) );
    exit( 1 );
  }
  return;
#endif
  fprintf( stderr, "Please define the test type\n" );
  exit( 1 );
} // End of openMem

inline void closeMem( int unmapPages ) {
  if ( memPtr == NULL && unmapPages != FALSE ) {
    fprintf( stderr, "ERROR: closeMem called without memory mapped (ignored)\n" );
  }
#ifdef PHYSMEM_TEST
  if ( memPtr != NULL ) {
    if ( unmapPages ) {
      if ( munmap( (void *)memPtr, currSizeBytes ) != 0 )
	perror( "Failed to unmap PHYSMEM (ignored)" );
    }
    memPtr = NULL;
  }
  if ( physmemFd != -1 ) {
    close( physmemFd );
    physmemFd = -1;
  } else {
    fprintf( stderr, "ERROR: closeMem called with physmem already opened (ignored)\n" );
  }
  return;
#endif
#ifdef IPC_TEST
  if ( memPtr != NULL ) {
    if ( shmdt( (void *)memPtr ) != 0 ) {
      perror( "Error during IPC detach (ignored)" );
    }
    memPtr = NULL;
  }
  if ( shmid != -1 ) {
    if ( shmctl( shmid, IPC_RMID, NULL ) != 0 ) {
      perror( "Error during IPC RM (ignored)" );
    }
    shmid = -1;
  }
  return;
#endif
  fprintf( stderr, "Please define the test type\n" );
  exit( 1 );
} // End of openMem
// ===========================================================

inline void getNow( struct timeval *nowIn ) {
  if ( nowIn == NULL )
    gettimeofday( &now, NULL );
  else
    gettimeofday( nowIn, NULL );
} // End of getNow

char *dumpThroughput( const double perByte ) {
  static char o[1024];

  const unsigned long long throughput = 1000000.0 / perByte;
  const unsigned long long kilo = 1024;
  if ( throughput < kilo ) {
    snprintf( SP(o), "%lld B/s", throughput );
  } else if ( throughput < kilo * kilo ) {
    snprintf( SP(o), "%.2f %sB/s", (double)throughput / kilo, "K" );
  } else if ( throughput < kilo * kilo * kilo ) {
    snprintf( SP(o), "%.2f %sB/s", (double)throughput / (kilo * kilo), "M" );
  } else if ( throughput < kilo * kilo * kilo * kilo) {
    snprintf( SP(o), "%.2f %sB/s", (double)throughput / (kilo * kilo * kilo), "G" );
  } else {
    snprintf( SP(o), "%.2f %sB/s", (double)throughput / (kilo * kilo * kilo * kilo), "T" );
  }
  return( o );
} // End of dumpThroughput

// Calculate delta time in usec
inline long long deltaTime( struct timeval *end, struct timeval *start ) {
  return( (end->tv_sec - start->tv_sec) * 1000000 + (end->tv_usec - start->tv_usec) );
} // End of deltaTime

// Dump delta time in ASCII format
char *dumpTime( double deltaTime ) {
  static char o[ 1024 ];
  if ( deltaTime > 10 ) {
    snprintf( SP(o), "%lld", (long long) deltaTime );
  } else {
    snprintf( SP(o), "%.2e", deltaTime );
  }
  snprintf( AP(o), " usec" );
  if ( deltaTime > 99999 ) {
    snprintf( AP(o), " (%.2f sec)", (double)deltaTime / 1000000.0 );
  }
  return( o );
} // End of dumpTime

inline void doReadTest( long long thisTestSizeBytes ) {
  long long b;
  wordType dummyIn;
  volatile wordType *memIn;
  long long targetWords = thisTestSizeBytes / sizeof( wordType );
  for ( memIn = memPtr, dummyIn = 0, b = 0; b != targetWords; b++ ) {
    dummyIn |= *memIn++;
  }
} // End of doReadTest

void measureRead( long long thisTestSizeBytes ) {
  long long elapsed;
  int nLoops;

  if ( TEST_WARMUP > 0 )
    for ( nLoops = 0; nLoops != TEST_WARMUP; nLoops++ )
      doReadTest( thisTestSizeBytes );
    
  for (nLoops = 0, getNow( &start );;) {
    nLoops++;
    doReadTest( thisTestSizeBytes );
        
    getNow( &end );

    // elapsed = deltaTime( &end, &start ); break;

    if ( (elapsed = deltaTime( &end, &start )) > TEST_MIN_DURATION
      && nLoops >= TEST_MIN_LOOPS ) break;
  }
  const double perLoop = (double)elapsed / nLoops;
  const double perByte = perLoop / thisTestSizeBytes;
  printf( " Read" );
  // printf( " size:%-16lld", thisTestSize );
  // printf( " time:%-30s", dumpTime( perLoop ) );
  printf( " perByte:%s", dumpTime( perByte ) );
  printf( " => %s", dumpThroughput( perByte ) );
} // End of measureRead

inline void doWriteTest( long long thisTestSizeBytes ) {
  long long b;
  volatile wordType *memOut = memPtr;
  wordType wordOut = 0xdeadbeefdeadbeef;
  long long targetSizeWords = thisTestSizeBytes / sizeof( wordType );
  for ( b = 0; b != targetSizeWords; b++ ) {
    *memOut++ = wordOut;
  }
} // End of doWriteTest

void measureWrite( long long thisTestSizeBytes ) {
  long long elapsed;
  int nLoops;

  if ( TEST_WARMUP > 0 )
    for ( nLoops = 0; nLoops != TEST_WARMUP; nLoops++ )
      doWriteTest( thisTestSizeBytes );
    
  for (nLoops = 0, getNow( &start );;) {
    nLoops++;
    doWriteTest( thisTestSizeBytes );
    getNow( &end );

    // elapsed = deltaTime( &end, &start ); break;

    if ( (elapsed = deltaTime( &end, &start )) > TEST_MIN_DURATION
      && nLoops >= TEST_MIN_LOOPS ) break;
  }
  const double perLoop = (double)elapsed / (double)nLoops;
  const double perByte = perLoop / thisTestSizeBytes;
  printf( " Write" );
  // printf( " size:%-16lld", thisTestSize );
  // printf( " time:%-30s", dumpTime( perLoop ) );
  printf( " perByte:%s", dumpTime( perByte ) );
  printf( " => %s", dumpThroughput( perByte ) );
} // End of measureWrite

const char *dumpSize( long long s, char suffix ) {
  static int i=0;
  static char outStr[ 10 ][ 1024 ];
  static char outVal[ 1024 ];

  snprintf( SP(outVal), "%lld %c", s, suffix );
  if ( s < 1024 ) {
  } else if ( (double)s < 1024 * 1024 ) {
    snprintf( AP(outVal), " = %.2f K%c", (double)s/1024.0, suffix );
  } else if ( (double)s < 1024 * 1024 * 1024 ) {
    snprintf( AP(outVal), " = %.2f M%c", (double)s/(1024.0*1024.0), suffix );
  } else if ( (double)s < (double)(1024.0 * 1024 * 1024 * 1024) ) {
    snprintf( AP(outVal), " = %.2f G%c", (double)s/(1024.0 * 1024.0 * 1024.0), suffix );
  } else {
    snprintf( AP(outVal), " = %.2f T%c", (double)s/(1024.0 * 1024.0 * 1024.0 * 1024.0), suffix );
  }
  
  i = (i + 1) % 10;
  strncpy( outStr[i], outVal, sizeof( outVal ) );
  return( outStr[i] );
} // End of dumpSize

int main( int argc, char **argv ) {
  int i;
  unsigned long long oldSizeBytes;

  printf( "%s starting\n", argv[0] );

  thisFile = argv[0];
  
  printf( "\tWord size:%lu", sizeof( wordType ) );

  openMem( 0 );
  closeMem( FALSE );

  printf( "\tMemory size:%s\n", dumpSize( memSizeBytes, 'B' ) );

  double testPerc1[] = { 1.199, 1.191, 0, 2, 5, 10, 50, 0 };
  for ( oldSizeBytes = 0, i = 0; testPerc1[i] != 0; i++ ) {
    unsigned long long currSizeBytes = memSizeBytes / testPerc1[i];

    if ( currSizeBytes == 0 || currSizeBytes == oldSizeBytes ) continue;
    oldSizeBytes = currSizeBytes;
    printf( "Map size:%-25s", dumpSize( currSizeBytes, 'B' ) );
    unsigned long long testSizeBytes = currSizeBytes;
    if ( testSizeBytes > MAX_TEST_SIZE_BYTES ) testSizeBytes = MAX_TEST_SIZE_BYTES;
    printf( "\tTest size:%-25s", dumpSize( testSizeBytes, 'B' ) );
    
    openMem( currSizeBytes );
    measureRead( testSizeBytes );
    measureWrite( testSizeBytes );
    closeMem( TRUE );
    printf( "\n" );
  }

  exit( 0 );
} // main
