/* phymem.c kernel module */
/* version 6.6 */

#ifndef __KERNEL__
#define __KERNEL__
#endif

#include <linux/module.h>
#include <linux/version.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include "physmem_version.h"
#include "physmem.h"

 /* File changed name */
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)) && (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)))
#include <linux/autoconf.h>
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33))
#include <generated/autoconf.h>
#else
#include <linux/config.h>
#endif

 /* These were included in mm.h up to version 2.6.23 */
#ifndef _LINUX_FS_H
#include <linux/fs.h>
#endif
#ifndef _LINUX_SCHED_H
#include <linux/sched.h>
#endif

/* Access e820 map */
#include <asm/e820.h>
#include <linux/kallsyms.h>
#include <linux/kprobes.h>

#define MIN_LINUX_SIZE 500
#define BUFFER_SIZE 512

/* globals */
static char *conf = "/etc/physmem.cfg";
long physmemSizeRequest[PHYSMEM_MAXDEVICE];
static int physmemMB[PHYSMEM_MAXDEVICE];
static int maxdevice = 2; /* default is physmem0 and physmem1 */
static u_long physmemSize[PHYSMEM_MAXDEVICE];
static u_long physmemStart[PHYSMEM_MAXDEVICE];
static u_long physmemEnd[PHYSMEM_MAXDEVICE];
static int physmemNumZones[PHYSMEM_MAXDEVICE];
static int *physmemMapZones[PHYSMEM_MAXDEVICE];
static struct memZoneStruct *memZones;
int numPhysZones = 1;
void *h_physaddr;
u_long h_size;
static int getFullSize = 0;

const struct e820map *e820_ptr = (void *)NULL;
const struct e820map *e820_saved_ptr = (void *)NULL;
int memmap = 0;

 /* Information about (possible) memory areas */
 /* Number of memory zones: */

MODULE_DESCRIPTION("PHYSMEM DRIVER");
MODULE_LICENSE("GPL");

// #define DEBUG
#ifdef DEBUG
#define PRINTK(a) printk a
#else
#define PRINTK(a) /**/
#endif

/* function prototypes */
 /* krealloc only avaible in > 2.6.22 kernels -> wrapper function */
 /* args: old pointer, new desired size, GFP flags, old size */
void *my_krealloc (const void *, size_t, gfp_t, size_t);
int read_config(void);
int get_cmd_info (void);
int get_mem_info (struct memZoneStruct** memInfoTable);
int get_bios_info(struct memZoneStruct **memInfoTable);

int init_module (void);
void cleanup_module (void);
static int physmem_open (struct inode * inode, struct file * file);
static int physmem_close (struct inode * inode, struct file * file);
static int read_fs (const void *useraddr, void *where, u_int size);
static int write_fs_val (const u_long value, void * useraddr, u_int size);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36))
static long physmem_unlocked_ioctl (struct file *file,
                        unsigned int cmd, unsigned long arg);
#else
static int physmem_ioctl (struct inode *inode, struct file *file,
			  unsigned int cmd, unsigned long arg);
#endif
static int physmem_mmap (struct file * file, struct vm_area_struct * vma);
static struct file_operations physmem_fops = {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36))
  .unlocked_ioctl =   physmem_unlocked_ioctl,
  .mmap =             physmem_mmap,
  .open =             physmem_open,
  .release =          physmem_close,
#else
  ioctl:   physmem_ioctl,
  mmap:	   physmem_mmap,
  open:	   physmem_open,
  release: physmem_close,
#endif
};

 /* Used when full mapping is used and software tries to access a page in the hole */
 /* Generates a SIGBUS error */
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,26))
struct page *mmap_error(int error_code, struct vm_area_struct *vma, unsigned long address)
{
  return NOPAGE_SIGBUS; /* send a SIGBUS error */
}
static struct vm_operations_struct mmap_error_ops = {
  nopage:  (void *)mmap_error,
};
#else
int mmap_error (struct vm_area_struct *vma, struct vm_fault *vmf)
{
  return VM_FAULT_SIGBUS; /* send a SIGBUS error */
}
static struct vm_operations_struct mmap_error_ops = {
  fault:   (void *)mmap_error,
};
#endif

/* loading module physmem */
int init_module (void)
{
  char textChar[8], textLine[100];
  u_char *tst_addr;
  u_char twr, trd;
  unsigned long long mem4gb;
  u_long running_start;
  u_long *remaining_size;
  u_long total_remaining_size;
  u_long useSize, totSize;
  int i, j, n_page;
  extern unsigned long num_physpages;
  int kmaj1, kmaj2, kmaj3;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0))
  int slc_major, slc_minor;
#endif
  int arch64 = 0; /* default: 32-bit machine */
  struct memZoneStruct *memInfo; /* info about physical memory zones, 
                                               read in /proc/iomem */
  int memInfoNumber; /* number of zones found in /proc/iomem */
  int cmdSize; /* size of Linux in boot command */

  remaining_size = kmalloc (1 * sizeof (u_long), GFP_KERNEL);
  for (i=0; i<PHYSMEM_MAXDEVICE; i++){
    physmemSize[i] = 0;
    physmemStart[i] = 0;
    physmemNumZones[i] = 1;
    physmemMapZones[i] = kmalloc (1 * sizeof (int), GFP_KERNEL);
    physmemMapZones[i][0] = 1; /* physmem0 uses first memory zone */
    physmemSizeRequest[i] = 0;
  }

   /* Default : physmem0 will be 96MB, the other will be as big as possible */
  physmemSizeRequest[0] = PHYSMEM_96MB;

  memZones = kmalloc (1 * sizeof (struct memZoneStruct), GFP_KERNEL);

   /* Are we on a 64-bit machine ? */
  if (sizeof (void *) > 4)
    arch64 = 1;

  /* Linux main version code */
  kmaj1 = LINUX_VERSION_CODE >> 16;
  kmaj2 = (LINUX_VERSION_CODE >> 8) & 0xff;
  kmaj3 = LINUX_VERSION_CODE & 0xff;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0))
  slc_major = RHEL_MAJOR;
  slc_minor = RHEL_MINOR;
#endif

  /* report the start of the loading */
  printk (KERN_INFO "physmem: Loading driver version " PHYSMEM_VERSION "\n");
  printk (KERN_INFO "physmem: Driver module compiled on %s at %s\n", 
          __DATE__, __TIME__);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0))
  printk (KERN_INFO "physmem: Linux version code = %d.%d.%d"
          " (SLC %d.%d)"
          ", instruction set = %d bit\n", 
          kmaj1, kmaj2, kmaj3,
          slc_major, slc_minor, 
          arch64 == 1 ? 64 : 32);   
#else
  printk (KERN_INFO "physmem: Linux version code = %d.%d.%d"
          ", instruction set = %d bit\n",
          kmaj1, kmaj2, kmaj3,
          arch64 == 1 ? 64 : 32);
#endif

  /* read configuration file and get physmem device sizes */
  maxdevice = read_config();

  /* steering the physmem size assignments with the module parameters 
     read on /etc/physmem.cfg 
 
     parameter physmemMB[i] (i = 0 ... 11): 
     >0  => request for this amount 
     Note: These are only requests, the actual size will be assigned 
     depending on the available total size of physmem.
     0   => request for 96 MB (default for /dev/physmem0)
     0   => request for the maximum avaiable size (default for physmem1 to 7)
     <0  => request for 0 size 
  */

  printk (KERN_INFO "physmem: Number of requested devices = %d\n", maxdevice);
  textLine[0] = '\0';
  for (i = 0; i < maxdevice; i++)
  {
    if (i < (maxdevice - 1))
      sprintf(textChar, "%d, ", physmemMB[i]);
    else
      sprintf(textChar, "%d\n", physmemMB[i]);
    strcat(textLine, textChar);
  }
  printk (KERN_INFO "physmem: Requested size of physmem0-%d = %s MB\n",
               maxdevice - 1, textLine);

  /* read and print Linux system memory for /proc/cmdline */
  cmdSize = get_cmd_info();
  if (cmdSize < 0)
    return -EIO;
  else if (cmdSize == 0)
    return -EPERM;
  printk (KERN_INFO "physmem: Memory specified in boot command: %dMB (0x%lx)\n", 
                                     cmdSize, (u_long)cmdSize << 20);

  /* determine memory end of Linux by using num_physpages or totalram_pages*/
  if (memmap)
  {
    h_physaddr = (void *) (totalram_pages * PAGE_SIZE);
    printk (KERN_INFO "physmem: Linux total size: %lu bytes (%d MB), %d pages\n",
      (u_long)h_physaddr, (int)((u_long)h_physaddr >> 20), (int)totalram_pages);
  }
  else
  {
    h_physaddr = (void *) (num_physpages * PAGE_SIZE);
    printk (KERN_INFO "physmem: Linux total size: %lu bytes (%d MB), %d pages\n",
      (u_long)h_physaddr, (int)((u_long)h_physaddr >> 20), (int)num_physpages);	
  }
  if ((u_long)h_physaddr < (MIN_LINUX_SIZE << 20))
  {
    printk (KERN_INFO "physmem: Too few memory assigned for Linux in /etc/grub.cfg\n"); 
    printk (KERN_INFO "physmem: At present Linux uses %ld MB, assign significantly more than %d MB\n", 
                      (u_long)h_physaddr >> 20, MIN_LINUX_SIZE);
    return -EPERM;
  }

  if (memmap)
  {
    if (cmdSize > (u_long)h_physaddr >> 20)
      h_physaddr = (void *) ((u_long)cmdSize << 20);
  }
  
  //Get info about memory from /proc/iomem 
  memInfo = kmalloc (sizeof (struct memZoneStruct), GFP_KERNEL);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
  memInfoNumber = get_mem_info (&memInfo);  // read /proc/iomem
#else
  memInfoNumber = get_bios_info (&memInfo); // via kallsyms
#endif
//printk (KERN_INFO "physmem: memInfoNumber = %d\n", memInfoNumber);
  if ((memInfoNumber < 1) && (arch64 == 0)&&
                             (cmdSize == ((u_long)h_physaddr >> 20)))
  {
    // determine memory size of physmem by scanning
    h_size = 0;
    n_page = 0;
    for (twr=1; ; twr++)
    {
      tst_addr = ioremap((u_long)(h_physaddr + h_size), PAGE_SIZE);
      if (tst_addr == NULL)
      {
        break;
      }
      writeb(twr, tst_addr);
      trd = twr -1;
      trd = readb(tst_addr);
      iounmap(tst_addr);
      if (trd == twr)
      {
        h_size += PAGE_SIZE;
        n_page++;
      }
      else
      {
//  printk (KERN_INFO "physmem: page# = %d: trd (%d) != twr(%d)\n",
//                                                       n_page, trd, twr);
        break;
      }
      schedule();
    } // for twr
    if (h_size > 0)
    {
      memInfoNumber = 1;
      memInfo[0].start = (u_long)h_physaddr;
      memInfo[0].end = (u_long)h_physaddr + h_size;
      memInfo[0].size = memInfo[0].end - memInfo[0].start;
    }
  } // memInfoNumber < 1 && arch32

  if (memInfoNumber < 1)
  {
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,30))
    printk (KERN_INFO "physmem: There is no free System RAM for physmem!\n");
    printk (KERN_INFO "physmem: Assign less memory for Linux in /etc/grub.cfg\n"); 
    printk (KERN_INFO "physmem: At present Linux uses %ld MB, assign significantly less\n", 
                      (u_long)h_physaddr >> 20);
#else
    printk (KERN_INFO "physmem: System table does not contain BIOS memory info!\n");
#endif
    return -EPERM;
  }

  /* register major device number */
  if (register_chrdev (PHYSMEM_MAJOR, PHYSMEM_NAME, &physmem_fops))
  {
    printk (KERN_INFO "physmem: unable to register major device number %d \n",
	    PHYSMEM_MAJOR);
    return -EIO;
  }

  /* Populate memZones array with size of different memory zones */
  numPhysZones = memInfoNumber;
  remaining_size = my_krealloc (remaining_size, numPhysZones * sizeof (u_long),
                                              GFP_KERNEL, 1 * sizeof (u_long));
  memZones = my_krealloc (memZones, numPhysZones * sizeof (struct memZoneStruct), 
                                GFP_KERNEL, 1 * sizeof (struct memZoneStruct));
  j = 0;
  for (i=0; i < memInfoNumber; i++)
  {
    mem4gb = MEM_4GB; // to make 32 bit cormlpiler happy 
    if ((i > 0) && ((unsigned long long)memInfo[i].start < mem4gb))
    {
      // Allow only 1 zones below 4GB
      numPhysZones--;
      continue;
    }
    memZones[j].start = memInfo[i].start;
    memZones[j].end = memInfo[i].end / PAGE_SIZE * PAGE_SIZE;
    memZones[j].size = memZones[j].end - memZones[j].start;
// printk (KERN_INFO "physmem: memZones[%d] start, end, size: 0x%lx, 0x%lx, 0x%lx\n", 
//                 j, memZones[j].start, memZones[j].end, memZones[j].size);
    j++;
  }

   /* Save the sizes of memory zones + total size of available memory */
  running_start = memZones[0].start;
  total_remaining_size = 0;
  for (i=0; i<numPhysZones; i++){
    remaining_size[i] = memZones[i].size;
    total_remaining_size += memZones[i].size;
  }
// printk (KERN_INFO "physmem: total_remaning : %#lx\n", total_remaining_size);

  for (i = 0; i < PHYSMEM_MAXDEVICE; i++)
  {
    if (physmemMB[i] > 0) {
      physmemSizeRequest[i] = (long)physmemMB[i] << 20; // MB to bytes
    } else if (physmemMB[i] < 0) {
      physmemSizeRequest[i] = -1;
    }
  }

   /* Populate the physmem devices, and save informations about number of used
          zones and mapping of zones to physmem devices */
  for (i=0; i<maxdevice; i++)
  {
    physmemStart[i] = running_start;
    physmemEnd[i]   = running_start;
    if ((physmemSizeRequest[i] != -1) &&
         physmemSizeRequest[i] > total_remaining_size)
      physmemSizeRequest[i] = total_remaining_size;
    if (physmemSizeRequest[i] <= total_remaining_size)
    {
      for (j=0; j<numPhysZones; j++)
      {
//      printk (KERN_INFO "physmem: device: %d, zone: %d\n", i, j);
        if (physmemSizeRequest[i] == -1)
          break; // go to next device 
        if (remaining_size[j] == 0)
          continue; // go to next physmemZone
        if (physmemSizeRequest[i] == 0)
        {
          physmemSizeRequest[i] = total_remaining_size;
        }
//      printk (KERN_INFO "physmem: Size: %#lx, Rqst: %#lx, remain: %#lx, start: %#lx\n",
//      physmemSize[i], physmemSizeRequest[i], remaining_size[j], running_start);
        if (physmemSizeRequest[i] <= remaining_size[j])
        {
//        printk (KERN_INFO "physmem: IF\n");
          physmemSize[i] += physmemSizeRequest[i];
          total_remaining_size -= physmemSizeRequest[i];
	  remaining_size[j] -= physmemSizeRequest[i];
          physmemMapZones[i][physmemNumZones[i]-1] = j+1;
          running_start += physmemSizeRequest[i];
          physmemEnd[i] = running_start;
          physmemSizeRequest[i] = -1;
        }
        else
        {
//        printk (KERN_INFO "physmem: ELSE\n");
	  physmemSize[i] += remaining_size[j];
          physmemSizeRequest[i] -= remaining_size[j];
          total_remaining_size -= remaining_size[j];
          remaining_size[j] = 0;
          physmemNumZones[i] = physmemNumZones[i]+1;
          physmemMapZones[i] = my_krealloc (physmemMapZones[i], physmemNumZones[i] * sizeof (int), GFP_KERNEL, (physmemNumZones[i]-1) * sizeof (int));
          physmemMapZones[i][physmemNumZones[i]-2] = j+1;
          running_start = memZones[physmemMapZones[i][physmemNumZones[i]-2]].start;
        }
//       printk (KERN_INFO "physmem: Size: %#lx, Rqst: %#lx, remain: %#lx, start: %#lx\n",
//       physmemSize[i], physmemSizeRequest[i], remaining_size[j], running_start);
      } // numPhysZones
    } // if (physmemSizeRequest[i] <= total_remaining_size)
  } // maxdevice

  /* Report some informations */
  printk (KERN_INFO "physmem: Linux ends at physical address: %#lx\n",
          (u_long)h_physaddr-1);
  useSize = 0;
  for (i = 0; i < numPhysZones; i++)
    useSize += memZones[i].size;
  printk (KERN_INFO "physmem: physmem usable size: %lu bytes (%lu MB), %lu pages\n",
          useSize, useSize >> 20, useSize/PAGE_SIZE);
//  totSize = memZones[numPhysZones-1].end - memZones[0].start;
  totSize = physmemEnd[maxdevice - 1] - physmemStart[0];
  printk(KERN_INFO "physmem: physmem total size: %lu (%#lx) bytes (%lu MB)\n",
          totSize, totSize, (u_long)(totSize >> 20));

  for (i = 0; i < maxdevice; i++) 
  {
    /* print start end and number of zones used by the device */
    textLine[0] = '\0';
    for (j = 0; j < physmemNumZones[i]; j++)
    {
      if (j < physmemNumZones[i] - 1) 
        sprintf (textChar, "[%d] ", physmemMapZones[i][j]-1);
      else
        sprintf (textChar, "[%d]\n", physmemMapZones[i][j]-1);
      strcat(textLine, textChar);
    }
    printk (KERN_INFO "physmem: Device physmem%d from %#lx to %#lx uses %d zone(s): %s\n", 
            i,  physmemStart[i], physmemEnd[i], physmemNumZones[i], textLine);

    /* print usable and total sizes of the device */
    if (i < (maxdevice - 1))
    {
      totSize = (physmemStart[i+1] - physmemStart[i]);
    }
    else
    {
//      totSize = memZones[numPhysZones-1].end - physmemStart[i];
      totSize = physmemEnd[i] - physmemStart[i];
    }
    if (physmemSize[i] == totSize)
      printk (KERN_INFO "physmem: device physmem%d usable size: %lu (%#lx) bytes (%lu MB)\n",
	    i, physmemSize[i], physmemSize[i], (u_long)(physmemSize[i] >> 20));
    else
    {
      printk (KERN_INFO "physmem: device physmem%d usable size: %lu (%#lx) bytes (%lu MB)\n",
	    i, physmemSize[i], physmemSize[i], (u_long)(physmemSize[i] >> 20));
      printk(KERN_INFO "physmem: device physmem%d total size: %lu (%#lx) bytes (%lu MB)\n", 
            i, totSize, totSize, (u_long)(totSize >> 20));
    }
  }

  for (i = 0; i < numPhysZones; i++)
  {
    printk (KERN_INFO "physmem: physZone%d start: %#lx end: %#lx, size: %ld MB\n", 
    i, memZones[i].start, memZones[i].end, (u_long)(memZones[i].size >> 20));
    if (i > 0)
      printk (KERN_INFO "physmem: Size of the memory gap: %ld bytes (%ld MB)\n",
                               memZones[i].start - memZones[i-1].end, 
                              (memZones[i].start - memZones[i-1].end) >> 20);  
    if (memZones[i].end <= memZones[i].start)
      printk (KERN_INFO "physmem: ATTENTION! memZones[%d].end <= memZones[%d].start\n", i, i);
  }

  kfree (memInfo); /* free the dynamically allocated table */
  
  // DEBUG ONLY
/*  for (i = 1; i< physmemNumZones[1]; i++)
  {
    printk (KERN_INFO "memZones[physmemMapZones[1][i-1]-1].start = %#lx\n", memZones[physmemMapZones[1][i-1]-1].start);
    printk (KERN_INFO "physmemStart[1] = %#lx\n", physmemStart[1]);
    printk (KERN_INFO "memZones[physmemMapZones[1][i-1]-1].size - physmemStart[1] = %#lx\n", memZones[physmemMapZones[1][i-1]-1].size - physmemStart[1]);
    printk (KERN_INFO "\n\n");
    printk (KERN_INFO "memZones[physmemMapZones[minor][i-1]-1].size - physmemSize[physmemMapZones[minor][i-1]-1] = %#lx\n", memZones[physmemMapZones[1][i-1]-1].size - physmemSize[physmemMapZones[1][i-1]-1]);
    printk (KERN_INFO "doit etre egal a : offset entre debut de nouvelle zone utilisable et debut de deivce\n");
    printk (KERN_INFO "non-continu : memZones[physmemMapZones[minor][i]-1].start - physmemStart[minor] = %#lx\n", memZones[physmemMapZones[1][i]-1].start - physmemStart[1]);
    printk (KERN_INFO "memZones.start = %#lx\n", memZones[physmemMapZones[1][i]-1].start);
    printk (KERN_INFO "physmemStart [minor] = %#lx\n", physmemStart[1]);
  }*/
  
  return 0;
}

 /* Unloading module physmem */
void cleanup_module (void)
{  
  int i;
   /* Unregister device */
  unregister_chrdev (PHYSMEM_MAJOR, PHYSMEM_NAME);
  h_physaddr = NULL;

   /* Free allocated memory */
  kfree (memZones);
  memZones = NULL;
  for (i=0; i<numPhysZones; i++){
    kfree (physmemMapZones[i]);
    physmemMapZones[i] = NULL;
  }

  printk(KERN_INFO "physmem: unloading driver version " PHYSMEM_VERSION "\n");  
}

 /* Open a physmem device */
static int physmem_open (struct inode * inode, struct file * file)
{
  long minor;

  minor = MINOR(inode->i_rdev);	
  if (minor >= maxdevice) {
    printk (KERN_INFO "physmem: unable to open minor device number %ld \n",
	    minor);
    return -EPERM;; // Operation not permitted
  }

  /* Remember minor device number for physmem_mmap */
  file->private_data = (void*)minor;

  /* MOD_INC_USE_COUNT; not used */

  return 0;
}

/* Close a physmem device */
static int physmem_close (struct inode * inode, struct file * file)
{
  long minor; 

  minor = MINOR(inode->i_rdev);	
  if (minor >= maxdevice) {
    printk (KERN_INFO "physmem: unable to close minor device number %ld \n",
	    minor);
    return -EPERM;;
  }

  /* MOD_DEC_USE_COUNT; not used */

  return 0;
}

/* mmap call on a physmem device */
static int physmem_mmap (struct file * file, struct vm_area_struct * vma)
{
  int err = 0;
  u_long size; // vma->vm_end - vma->vm_start;
  u_long requestedSize, allocatedSize, gapSize;
  u_long offset;
  u_long virt_offset; // used to calculate the virtual offset of the different zones
  long minor;
  int i;

  requestedSize = vma->vm_end - vma->vm_start;
  size = requestedSize;
  allocatedSize = 0;
  minor = (long)(file->private_data);
  if (minor >= maxdevice) {
    printk (KERN_INFO "physmem: unable to mmap minor device number %ld \n",
	    minor);
    return -EPERM;;
  }

  offset = physmemStart[minor];
  vma->vm_pgoff = offset >> PAGE_SHIFT;
  virt_offset = 0;
 
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32))
  // Set the appropriate vm_flags:
  // DONTCOPY: do not copy the pages to children
  // NODUMP: do not include the pages in the core dumps
  vma->vm_flags |= VM_DONTCOPY;
  vma->vm_flags |= VM_NODUMP;
#endif

  if( (offset|size) & ~PAGE_MASK )
    return -EFAULT;

  PRINTK((KERN_EMERG "physmem: minor=%ld vm_start=0x%lx offset=0x%lx size=0x%lx getFullSize = %d\n",
	  minor, vma->vm_start, offset, size, getFullSize));

  /* remap_page_range inside mm/memory.c depends on the kernel version */
  /* doesn't exist in newer kernels, replaced by remap_pfn_range */
  vma->vm_ops = &mmap_error_ops;
  /* map contiguously or not, depending on what ioctl was called 
                                                     (GETSIZE or GETFULLSIZE) */
  if (getFullSize == 0) /* GETSIZE was used 
                         - map contiguously the different parts of the device */
  {
     /* if the device uses more than one memory zone, 
                               map them contiguously in virtual address space */
    if (physmemNumZones[minor] > 1)
    {
      /* first memory zone might not be completely used */
      size =  memZones[physmemMapZones[minor][0]-1].end - physmemStart[minor];
      if (requestedSize < size)
        size = requestedSize;
      PRINTK ((KERN_INFO "physmem: first GETSIZE offset: %#lx, size: %#lx\n", 
                                                                 offset, size));
      err=remap_pfn_range(vma, vma->vm_start, offset>>PAGE_SHIFT, size, 
                                                       vma->vm_page_prot);
      virt_offset = memZones[physmemMapZones[minor][0]-1].end - 
                             physmemStart[minor];
      allocatedSize += size;

      /* all memory zones completely used (between first and last) */
      for (i=1; (allocatedSize < requestedSize) && 
                (i < physmemNumZones[minor]-1); i++)
      {
        offset = memZones[physmemMapZones[minor][i]-1].start;
        size = memZones[physmemMapZones[minor][i]-1].size;
        if (requestedSize < allocatedSize + size);
          size = requestedSize - allocatedSize;
        PRINTK ((KERN_INFO "physmem: GETSIZE zone: %d offset: %#lx, size: %#lx, virt_offset: %#lx\n", i, offset, size, virt_offset));
        err=remap_pfn_range(vma, vma->vm_start + virt_offset, 
                                offset>>PAGE_SHIFT, size, vma->vm_page_prot);
        // virtual offset of zone = sum of all other zone sizes
        virt_offset += memZones[physmemMapZones[minor][i]-1].size;
        allocatedSize += size;
      }

      /* last memory zone might not be completely used */
      if (allocatedSize < requestedSize)
      {
        i = physmemNumZones[minor]-1;
        offset = memZones[physmemMapZones[minor][i]-1].start;
        size = physmemEnd[minor] - memZones[physmemMapZones[minor][i]-1].start;
        if (requestedSize < allocatedSize + size);
          size = requestedSize - allocatedSize;
        PRINTK ((KERN_INFO "physmem: last GETSIZE zone: %d offset: %#lx size: %#lx virt_offset: %#lx\n", i, offset, size, virt_offset));
        err=remap_pfn_range(vma, vma->vm_start + virt_offset, 
                                offset>>PAGE_SHIFT, size, vma->vm_page_prot);
        allocatedSize += size;
      }
    }
    else // only one Zone 
    {
      PRINTK((KERN_INFO "physmem: GETSIZE offset: %#lx, size: %#lx\n",
                                                               offset, size));
      PRINTK((KERN_INFO "physmem: size:%lu prot:%08lx vm_start:%lx pfn:%ld\n",
            size, vma->vm_page_prot.pgprot, vma->vm_start, offset>>PAGE_SHIFT));
      PRINTK((KERN_INFO "physmem: before flags:%08lx prot:%08lx\n", 
                                      vma->vm_flags, vma->vm_page_prot.pgprot));
      err = remap_pfn_range(vma, vma->vm_start, offset>>PAGE_SHIFT, size, 
                                                            vma->vm_page_prot);
      PRINTK((KERN_INFO "physmem: after  flags:%08lx prot:%08lx\n", 
                                      vma->vm_flags, vma->vm_page_prot.pgprot));
      allocatedSize += size;
    }
  }
  else /* if (getFullSize == 1) GETFULLSIZE was called, map not contiguously 
                             - physical hole is visible as a "virtual hole" */
  {
    if (physmemNumZones[minor] > 1)
    {
      /* first memory zone might not be completely used */
      size =  memZones[physmemMapZones[minor][0]-1].end - physmemStart[minor];
      if (requestedSize < size)
        size = requestedSize;
      PRINTK((KERN_EMERG "physmem: first FULLGETSIZE offset: %#lx, size: %#lx\n",
                                                                offset, size));
      err=remap_pfn_range(vma, vma->vm_start, offset>>PAGE_SHIFT, size, 
                                                            vma->vm_page_prot);
      gapSize = memZones[physmemMapZones[minor][0]].start - 
                memZones[physmemMapZones[minor][0]-1].end;
      allocatedSize += size + gapSize;

      /* all memory zones completely used (between first and last) */
      for (i=1; (allocatedSize < requestedSize) && 
                (i < physmemNumZones[minor]-1); i++){
        offset = memZones[physmemMapZones[minor][i]-1].start;
        size = memZones[physmemMapZones[minor][i]-1].size;
        if (requestedSize < allocatedSize + size);
          size = requestedSize - allocatedSize;
        virt_offset = memZones[physmemMapZones[minor][i]-1].start - 
                                                            physmemStart[minor];
        PRINTK ((KERN_EMERG "physmem: FULLGETSIZE zone: %d offset: %#lx, size: %#lx, virt_offset: %#lx\n", i, offset, size, virt_offset));
        err=remap_pfn_range(vma, vma->vm_start + virt_offset, 
                                  offset>>PAGE_SHIFT, size, vma->vm_page_prot);
        gapSize = memZones[physmemMapZones[minor][i]].start - 
                  memZones[physmemMapZones[minor][i]-1].end;
        allocatedSize += size + gapSize;
      }

      /* last memory zone might not be completely used */
      if (allocatedSize < requestedSize)
      {
        i = physmemNumZones[minor]-1;
        offset = memZones[physmemMapZones[minor][i]-1].start;
        size = physmemEnd[minor] - memZones[physmemMapZones[minor][i]-1].start;
        if (requestedSize < allocatedSize + size);
          size = requestedSize - allocatedSize;
        virt_offset = memZones[physmemMapZones[minor][i]-1].start - 
                                                            physmemStart[minor];
        PRINTK ((KERN_EMERG "physmem: last FULLGETSIZE zone: %d offset: %#lx, size: %#lx, virt_offset: %#lx\n", i, offset, size, virt_offset));
        err=remap_pfn_range(vma, vma->vm_start + virt_offset, 
                                   offset>>PAGE_SHIFT, size, vma->vm_page_prot);
        allocatedSize += size;
      }
    }
    else // only one Zone
    {
    PRINTK((KERN_INFO "physmem: FULLGETSIZE offset: %#lx, size: %#lx\n",
                                                               offset, size));
      err=remap_pfn_range(vma, vma->vm_start, offset>>PAGE_SHIFT, size, 
                                                            vma->vm_page_prot);
      allocatedSize += size;
    }
  }

  PRINTK((KERN_EMERG "physmem: minor=%ld allocatedSize=0x%lx err=%d\n",
                                                   minor, allocatedSize, err));
  if (err) return err;
 
  return 0;
}

/* ioctl call on a physmem device */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36))
static long physmem_unlocked_ioctl (struct file *file,
                        unsigned int cmd, unsigned long arg)
#else
static int physmem_ioctl (struct inode *inode, struct file *file,
			  unsigned int cmd, unsigned long arg)
#endif
{
  int ret=0;  
  u_short dir;
  u_int arg_size;
  u_long arg_val;
  u_long beginning;
  u_long end;
  u_long size;
  int minor;
  int i;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36))
  minor = (long)(file->private_data);
#else
  minor = MINOR(inode->i_rdev);	
#endif
  if (minor >= maxdevice) {
    printk (KERN_INFO "physmem: unable to ioctl minor device number %d \n",
	    minor);
    return -EPERM;;
  }

  dir = _IOC_DIR(cmd);
  arg_size = _IOC_SIZE(cmd);
  if ( (dir == _IOC_WRITE) && arg_size && (arg_size <= sizeof(u_long)) )
    if ( (ret = read_fs((const void *)arg, &arg_val, arg_size)) ) {
      return ret;
    }

   /* return the requested values */
  switch (cmd) {

   /* get physical start address of device */
  case PHYSMEM_GETADDR:
    ret = write_fs_val (physmemStart[minor], (void *)arg, arg_size);
    if (ret){
      printk (KERN_INFO "physmem: error: wrong argument in ioctl call (PHYSMEM_GETADDR)\n");
      return ret;
    }
    break;

   /* get usable size of device */
  case PHYSMEM_GETSIZE:
    ret = write_fs_val (physmemSize[minor], (void *)arg, arg_size);
    if (ret){
      printk (KERN_INFO "physmem: error: wrong argument in ioctl call (PHYSMEM_GETSIZE)\n");
      return ret;
    }
    getFullSize = 0;
    break;

   /* get full size of device, including holes */
  case PHYSMEM_GETFULLSIZE:

     /* start */
    if (physmemStart[minor] >= memZones[physmemMapZones[minor][0]-1].start){
      beginning = physmemStart[minor];
    }
    else{
      beginning = memZones[physmemMapZones[minor][0]-1].start;
// check
    }

     /* end */
    if (physmemEnd[minor] <= memZones[physmemMapZones[minor][physmemNumZones[minor]-1]-1].end){
      end = physmemEnd[minor];
    }
    else {
      end = memZones[physmemMapZones[minor][physmemNumZones[minor]-1]-1].end;
// check
    }

     /* size */
    size = end - beginning;
    ret = write_fs_val (size, (void *)arg, arg_size);
    if (ret){
      printk (KERN_INFO "physmem: error: wrong argument in ioctl call (PHYSMEM_GETFULLSIZE)\n");
      return ret;
    }
    getFullSize = 1;
    break;

   /* how many zones are used by device */
  case PHYSMEM_GETNUMMEMZONES:
    ret = write_fs_val (physmemNumZones[minor], (void *)arg, arg_size);
    if (ret){
      printk (KERN_INFO "physmem: error: wrong argument in ioctl call (PHYSMEM_GETNUMMEMZONES)\n");
    }
    break;

   /* which zones are used by device (beginning-end addresses + size) */
   /* to be called with an array of PHYSMEM_GETMEMZONES elements of type struct memZoneStruct */
  case PHYSMEM_GETMEMZONES:
     /* received a pointer to a struct -> use it as an array */
    ret = 0;
    for (i=0; i<physmemNumZones[minor]; i++){

       /* start */
      if (physmemStart[minor] >= memZones[physmemMapZones[minor][i]-1].start){
        beginning = physmemStart[minor];
      }
      else{
        beginning = memZones[physmemMapZones[minor][i]-1].start;
      }
      ret = write_fs_val (beginning, &(((struct memZoneStruct *) arg)[i].start), sizeof (beginning));
      if (ret){
        printk (KERN_INFO "physmem: error: wrong argument in ioctl call (PHYSMEM_GETMEMZONES)\n");
        return ret;
      }

       /* end */
      if (physmemEnd[minor] <= memZones[physmemMapZones[minor][i]-1].end){
        end = physmemEnd[minor];
      }
      else {
        end = memZones[physmemMapZones[minor][i]-1].end;
      }
      ret = write_fs_val (end, &(((struct memZoneStruct *) arg)[i].end), sizeof (end));
      if (ret){
        printk (KERN_INFO "physmem: error: wrong argument in ioctl call (PHYSMEM_GETMEMZONES)\n");
        return ret;
      }

       /* size */
      size = end - beginning;
      ret = write_fs_val (size, &(((struct memZoneStruct *) arg)[i].size), sizeof (size));
      if (ret){
        printk (KERN_INFO "physmem: error: wrong argument in ioctl call (PHYSMEM_GETMEMZONES)\n");
        return ret;
      }
    }
    break;

  default:
    printk (KERN_INFO "physmem: invalid ioctl request %d\n",cmd);
    return -ENOTTY;
  }  
  return 0;
}

/* read_fs */
static int read_fs (const void *useraddr, void *where, u_int size)
{
  int err;
 
  if (useraddr == NULL || !size)
    return -EINVAL;

  /* function verify_area changed to access_ok in > 2.6.14 kernels */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14))
  if ((err = !access_ok(VERIFY_READ, useraddr, size)))
#else
  if ((err = verify_area(VERIFY_READ, useraddr, size)))
#endif
    return err;
  switch (size) {
  case 1:
    get_user (*(u_char *)where, (u_char *)useraddr);
    break;
  case 2:
    get_user (*(u_short *)where, (u_short *)useraddr);
    break;

   /* u_long is 4 bytes wide on 32bit, 8 bytes wide on 64 bit machines */
  case 4:
  case 8:
    get_user (*(u_long *)where, (u_long *)useraddr);
    break;
  default:
    err = copy_from_user (where, useraddr, size);
  }
  return 0;
}

/* write_fs */
static int write_fs_val (const u_long value, void * useraddr, u_int size)
{
  int err;
 
  if (useraddr == NULL)
    return -EINVAL;

  /* function verify_area changed to access_ok in > 2.6.14 kernels */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14))
  if ((err = !access_ok (VERIFY_WRITE, useraddr, size)))
#else
  if ((err = verify_area (VERIFY_WRITE, useraddr, size)))
#endif
  {
    printk (KERN_INFO "address not writeable ! \n");
    return err;
  }

  switch (size) {
  case 1:
    err = put_user ((u_char)value, (u_char *)useraddr);
    break;
  case 2:
    err = put_user ((u_short)value, (u_short *)useraddr);
    break;
  case 4:
    err = put_user ((u_int)value, (u_int *)useraddr);
    break;
  case 8:
    err = put_user ((u_long)value, (u_long *)useraddr);
    break;
  default:
    return -EINVAL;
  }
  return 0;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22))

 /* krealloc was first introduces in 2.6.22 kernel */
 /* create a wrapper function to use it anyway - tested on 2.6.9 to 2.6.26 kernels */
void *my_krealloc (const void *p, size_t new_size, gfp_t flags, size_t old_size)
{
  void *ret;

  if (unlikely (!p))
    return kmalloc_track_caller(new_size, flags);

  if (unlikely (!new_size)){
    kfree (p);
    return NULL;
  }

  ret = kmalloc_track_caller (new_size, flags);
  if (!ret)
    return NULL;

  memcpy (ret, p, min (new_size, old_size));
  kfree (p);
  return ret;
}

#else /* kernel > 2.6.22: krealloc available -> let's use it */

void *my_krealloc (const void *p, size_t new_size, gfp_t flags, size_t old_size)
{
  return krealloc (p, new_size, flags);
}
#endif


int read_config(void)
{
  int posFile=0;
  int posLine=0;
  struct file *fd;
  char buff [1];
  char buffLine [BUFFER_SIZE];
  int size;
  int comLine=0;
  int i;

  /* set default values */
  size = 2;
  for (i = 0; i < PHYSMEM_MAXDEVICE; i++)
  {
    physmemMB[i] = 0;
  } 

  /* open configuration file to get information about device sizes */
  fd = filp_open (conf, O_RDONLY, 0);
  if (IS_ERR(fd)){
    printk (KERN_INFO "physmem: Cannot open %s! Default values are used.\n", conf);
    return size;
  }
  printk (KERN_INFO "physmem: Reading configuration file %s\n", conf);

  while (kernel_read (fd, posFile, buff, 1) == 1)
  {
    posFile++;
    buffLine[posLine] = buff[0];
    posLine++;

    if (posLine >= BUFFER_SIZE)
    {
      printk (KERN_INFO "physmem: %s is too long! Default values are used.\n", conf);
      break;
    }
 
    if (buff[0] == '\n')
    {
      buffLine[posLine] = '\0';
      if (buffLine[0] == '#')
      {
        posLine = 0;
        comLine++;
        continue;
      }
      printk (KERN_INFO "physmem: %s contains: %s\n", conf, buffLine);
      size = sscanf(buffLine, "%d %d %d %d %d %d %d %d %d %d %d %d", 
                                                         &physmemMB[0],
                                                         &physmemMB[1],
                                                         &physmemMB[2],
                                                         &physmemMB[3],
                                                         &physmemMB[4],
                                                         &physmemMB[5],
                                                         &physmemMB[6],
                                                         &physmemMB[7],
                                                         &physmemMB[8],
                                                         &physmemMB[9],
                                                         &physmemMB[10],
                                                         &physmemMB[11]);
      if (size < 2)
        size = 2;
      else if ((size < PHYSMEM_MAXDEVICE) && (physmemMB[size-1] != 0))
      {
        physmemMB[size] = 0;
        size++;
      }

      break;
    }
  }

  filp_close (fd, NULL);

  return size;
}

int get_cmd_info (void)
{
  int posLine=0;
  struct file *fd;
  char buff [1];
  char buffLine [BUFFER_SIZE];
  char c;
  char *memPos, *memmapPos;
  int size=0;
  int linuxBootSize=0;

  /* open /proc/cmdline to get information about system memory */
  fd = filp_open ("/proc/cmdline", O_RDONLY, 0);
  if (IS_ERR(fd)){
    printk (KERN_INFO "physmem: Cannot open /proc/cmdline !\n");
    return -1;
  }

  while (kernel_read (fd, posLine, buff, 1) == 1)
  {
    buffLine[posLine] = buff[0];
    posLine++;

    if (posLine >= BUFFER_SIZE)
    {
      printk (KERN_INFO "physmem: /proc/cmdline is too long!\n");
      break;
    }
 
    if (buff[0] == '\n')
    {
      buffLine[posLine] = '\0';
      printk (KERN_INFO "physmem: /proc/cmdline = %s\n", buffLine);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,32))
#ifdef CHECK_NOPAT
      /* try to find 'nopat' in command line */
      memPos = strstr(buffLine, "nopat");
      if (memPos == NULL)
      {
        printk (KERN_INFO 
          "physmem: No 'nopat' option specified in boot command!\n");
        break;
      }
#endif
#endif
      /* try to find 'mem=' in command line */
      memPos = strstr(buffLine, "mem=");
      if (memPos == NULL)
      {
        /* try to find 'memmap' in command line if 'mem=' not found */
        memmapPos = strstr(buffLine, "memmap=");
        if (memmapPos != NULL)
        {
          memmap = 1;
          memPos = strstr(memmapPos, "$");
        }
      }
      if (memPos == NULL)
      {
        printk (KERN_INFO 
          "physmem: No memory size specified in boot command!\n");
        break;
      }
      if (memmap)
      {
//      printk (KERN_INFO "physmem: Memory specified in boot command = %s\n", 
//                                                                  memPos+1);
        size = sscanf(memPos+1, "%d%cM", &linuxBootSize, &c);
      }
      else
      {
//      printk (KERN_INFO "physmem: Memory specified in boot command = %s\n", 
//                                                                  memPos+4);
        size = sscanf(memPos+4, "%d%cM", &linuxBootSize, &c);
      }
      if (size >  0)
      {
        if ((c == 'G') || (c == 'g'))
          size = linuxBootSize * 1024;
        else
          size = linuxBootSize;
      }
    }
  }
  filp_close (fd, NULL);

  return size;
}
int get_mem_info (struct memZoneStruct **memInfoTable)
{
  int posFile=0;
  int posLine=0;
  int lineNo=0;
  struct file *fd;
  char buff [1];
  char buffLine [BUFFER_SIZE];
  char buffStr1 [BUFFER_SIZE/4];
  char buffStr2 [BUFFER_SIZE/4];
  char c;
  long start, end;
  int size=1;
  
   /* open /proc/iomem to get information about system memory */
  fd = filp_open ("/proc/iomem", O_RDONLY, 0);
  if (IS_ERR(fd)){
    printk (KERN_INFO "physmem: Cannot open /proc/iomem !\n");
    return -1;
  }
  
//  i = 0;
  start = 0;
  end = 0;
  
   /* read the /proc/iomem file to get information */
  while (kernel_read (fd, posFile, buff, 1) == 1)
  {
    posFile++;
    
    buffLine[posLine] = buff[0];
    posLine++;
    
    if (buff[0] == '\n')
    {
      buffLine[posLine] = '\0';
      lineNo++;
//      printk (KERN_INFO "physmem: iomem line %d: %s\n", lineNo, buffLine);
      if (sscanf (buffLine, "%lx %c %lx %*s %s %s", 
                             &start, &c, &end, buffStr1, buffStr2) == 5)
      {
        if ((strcmp (buffStr1, "System") == 0) && 
            (strcmp (buffStr2, "RAM") == 0) &&
            (end > (u_long)h_physaddr))
        {
//          printk (KERN_INFO "physmem: System RAM %d, start: 0x%lx, end: 0x%lx\n",
//                                          size, start, end);
	        if (size > 1)
	          *memInfoTable = my_krealloc (*memInfoTable, 
                                        (size)*sizeof(struct memZoneStruct), 
                                        GFP_KERNEL,
                                        (size+1)*sizeof(struct memZoneStruct));
          if (start < (u_long)h_physaddr)
            (*memInfoTable)[size-1].start = (u_long)h_physaddr;
          else
            (*memInfoTable)[size-1].start = start;
	        (*memInfoTable)[size-1].end = end+1;
	        (*memInfoTable)[size-1].size = end+1 - (*memInfoTable)[size-1].start;
          // printk (KERN_INFO "physmem: System RAM %d, start: 0x%lx, end: 0x%lx, size: 0x%lx\n",
          //   size, (*memInfoTable)[size-1].start, end, (*memInfoTable)[size-1].size);
	        size++;
        }
      }
      /* 0 all data */
      memset (buffLine, 0, BUFFER_SIZE);
      memset (buff, 0, 1); 
      memset (buffStr1, 0, BUFFER_SIZE/4);
      memset (buffStr2, 0, BUFFER_SIZE/4);
      posLine = 0;
    }
  }
  filp_close (fd, NULL);
  
  return (size - 1);
}

int get_bios_info(struct memZoneStruct **memInfoTable)
{
  long start, end;
  int i;

  struct kprobe kp;
  void *(*my_kallsyms_lookup_name)(const char *name);
  const struct e820map* e820; 

  int size=1;

  /* Find the e820 addresses via the lookup system function */
  my_kallsyms_lookup_name = NULL;
  memset(&kp, 0, sizeof(kp));
  kp.symbol_name = "kallsyms_lookup_name";
  if (!register_kprobe(&kp))
  {
    my_kallsyms_lookup_name = (void *) kp.addr;
    unregister_kprobe(&kp);
  }
  if ( my_kallsyms_lookup_name == NULL ) {
    printk (KERN_ERR "physmem: Failed to lookup for kallsyms_lookup_name" );
    return -1;
  }
  e820_ptr = my_kallsyms_lookup_name( "e820" );
  e820_saved_ptr = my_kallsyms_lookup_name( "e820_saved" );
  if ( e820_ptr == NULL || e820_saved_ptr == NULL ) {
    printk (KERN_ERR "physmem: Failed to lookup for e820 structures" );
    return -1;
  }
  // printk (KERN_INFO "physmem: BIOS table via kallsyms_lookup_name e820:%p\n", 
  //                                            e820_ptr);
  // printk (KERN_INFO "physmem:                               e820_saved:%p\n",
  //                                      e820_saved_ptr);

  /* read the e820_saved table to get information */
  e820 = e820_saved_ptr;
  start = 0;
  end = 0;
  // printk (KERN_INFO "physmem: e820->nr_map=%d\n", e820->nr_map);

  for (i = 0; i < e820->nr_map; i++)
  {
    if (e820->map[i].type == 1)
    {
      // printk (KERN_INFO "physmem: e820.map[%d]->addr=0x%llx, size=0x%llx, type=%d\n",
      //     i, e820->map[i].addr, e820->map[i].size, e820->map[i].type);
      start = e820->map[i].addr;   
      end =   e820->map[i].addr + e820->map[i].size;
      if ((end-1) > (u_long)h_physaddr)
      {
        // printk (KERN_INFO "physmem: usable RAM %d, start: 0x%lx, end: 0x%lx\n",
        //              size, start, end);
        if (size > 1)
          *memInfoTable = my_krealloc (*memInfoTable,
                                    (size)*sizeof(struct memZoneStruct),
                                     GFP_KERNEL,
                                     (size+1)*sizeof(struct memZoneStruct));
        if (start < (u_long)h_physaddr)
          (*memInfoTable)[size-1].start = (u_long)h_physaddr;
        else
          (*memInfoTable)[size-1].start = start;
        (*memInfoTable)[size-1].end = end;
        (*memInfoTable)[size-1].size = end - (*memInfoTable)[size-1].start;
        // printk (KERN_INFO "physmem: System RAM %d, start: 0x%lx, end: 0x%lx, size: 0x%lx\n",
        //   size, (*memInfoTable)[size-1].start, end, (*memInfoTable)[size-1].size);
        size++;
      }
    }
  }

  return (size - 1);
}

