/*
**********************************************************************
**
**  physmemTest_mod.c
**  09/08/2012 by Gabor Kiss
**  according to:
**  utility to test the physmem memory selected physmem device
**  last modification 27-Apr-2013
**  by Klaus Schossmaier, Marc Fauman and Ervin Denes
**
**********************************************************************
*/

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include "physmem.h"
#include <time.h>

#ifndef datePointer
#define datePointer long
#endif

#define PATTERN 2

void write_physmem (void*, long, int);
void read_physmem (void*, long, int);
void usage(void)
{
  printf("GETSIZE test: open, write and read in physmem devices\n");
  printf("USAGE: physmemTest_mod [{-s|-S} <first physmem minor to test> | 0]\n"); 
  printf("                       [{-m|-M} <last physmem minor to test> | %d]\n",
                                                         PHYSMEM_MAXDEVICE);
  printf("                       [{-l|-L} <test length> | 0]\n");
  printf("                       [{-v|-V}]\n");
  printf("       -L 0 means: write and read to the whole physmem\n");
  printf("       -V   means verbose mode\n");
  exit(EXIT_FAILURE);
}

/*----------------------------------------------------------*/
int main(int argc, char **argv)
{
  int c, minor, physmem_fd;
  void *addr_user_physmem;
  unsigned long phys_addr_physmem;
  unsigned long phys_size_physmem;
  unsigned int phys_num_zones;
  unsigned long tlength, tlengthuser;
  struct memZoneStruct *phys_zones;
  int j;
  char physmem_name[16];
  int testm;
  int startminor=0;
  int maxminor = PHYSMEM_MAXDEVICE -1;
  int verbose = 0;

  clock_t start,end;
  double diftime;
//  double time_per_size;
  double MBperSec ;

  start=0;
  end=-1;

  testm = 1;
  minor = 0;
  tlengthuser=0;
  if(argc > 1)
  {
    while((c = getopt(argc, argv, "M:m:S:s:VvL:l:")) != EOF)
      switch(c)
      {
        case 'M':
        case 'm': maxminor = atoi(optarg); break;
        case 'S' :
        case 's': startminor = atoi(optarg); break;
        case 'L':
        case 'l': tlengthuser = atol(optarg); break;
        case 'V':
        case 'v': verbose = 1; break;
        default : usage();                break;
      }
  }
  tlength = tlengthuser;

  if ((maxminor < 0) || (maxminor >= PHYSMEM_MAXDEVICE))
  {
    printf("Maximum physmem minor %d is not valid\n", maxminor);
    usage();
  }
  if ((startminor < 0) || (startminor > maxminor))
  {
    printf("Starting physmem minor %d is not valid\n", startminor);
    usage();
  }
//  printf("# \t Minor \t time[s] \t physmemSize \t testLength  \t us/Byte\n");
  printf("# \t Minor \t time[s] \t physmemSize \t testLength  \t MByte/s\n");

  for (minor = startminor; minor <= maxminor; minor++)
  {
    if (verbose) printf("\nMinor %d:\n", minor);

    sprintf(physmem_name, "/dev/physmem%d", minor);
    if ((physmem_fd = open (physmem_name, O_RDWR)) == -1)
    {
      perror("open");
      printf("Cannot open device %s \n", physmem_name);
      exit (EXIT_FAILURE);
    }
    if (verbose) printf ("Opening physmem device %s\n", physmem_name);

    if (ioctl (physmem_fd, PHYSMEM_GETADDR, &phys_addr_physmem) < 0) {
      perror("ioctl");
      printf ("Cannot ioctl %s to get address\n", physmem_name);
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }
    if (verbose) printf ("Physical address = %#lx\n", phys_addr_physmem);
    if (verbose) fflush (NULL);

    /* get and print zones */
    if (ioctl (physmem_fd, PHYSMEM_GETNUMMEMZONES, &phys_num_zones) < 0) {
      perror("ioctl");
      printf ("Cannot ioctl device %s to get number of useable blocks\n",
                               physmem_name);
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }  
    if (verbose) printf (" + Number of memory zones : %d\n", phys_num_zones);
    if (verbose) fflush (NULL);

    phys_zones = (struct memZoneStruct *) malloc (phys_num_zones*sizeof (struct memZoneStruct));

    if (ioctl (physmem_fd, PHYSMEM_GETMEMZONES, phys_zones) < 0) {
      perror("ioctl");
      printf ("Cannot ioctl device %s to get memory zones \n", physmem_name);
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }

    for (j=0; j < phys_num_zones; j++)
    {
      if (verbose) printf (" + Mem zone %d : %#lx -> %#lx (size = %#lx)\n", 
        j+1, phys_zones[j].start, phys_zones[j].end, phys_zones[j].size);
    }
    if (verbose) fflush (NULL);

    /* only no hole test */
    if (ioctl (physmem_fd, PHYSMEM_GETSIZE, &phys_size_physmem) < 0){
      perror("ioctl");
      printf ("Cannot ioctl %s to get size\n", physmem_name);
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }
    if (verbose) printf ("Physical usable size of device (no hole) : %#lx\n", 
                                                  phys_size_physmem);
    if (verbose) fflush (NULL);

    if (phys_size_physmem <= 0)
    {
      printf ("Can not test 0 size\n");
      close (physmem_fd);
      continue;
    }
    
    if ((tlength==0) || (tlength > phys_size_physmem)) {
      tlength = phys_size_physmem;
    }
//    start=clock();
    if (verbose) printf ("mapp'ing\n");
    if (verbose) fflush (NULL);
    if ((addr_user_physmem = mmap (0, tlength, PROT_READ|PROT_WRITE, 
                    MAP_SHARED, physmem_fd, 0)) == MAP_FAILED){
      perror ("mmap");
      printf ("Mmap failed\n");
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }
    if (verbose) printf ("Mmap done to %p -> %p, trying to access physmem\n", 
    addr_user_physmem, addr_user_physmem+phys_size_physmem);
    if (verbose) fflush (NULL);

    start=clock();
    write_physmem (addr_user_physmem, tlength, verbose);
    read_physmem (addr_user_physmem, tlength, verbose);
    end=clock();

    if (verbose) printf ("unmapp'ing\n");
    if (verbose) fflush (NULL);
    munmap (addr_user_physmem, tlength);
//    end=clock();

    diftime = ((double) (end - start) / CLOCKS_PER_SEC);
//    time_per_size = 1000000.0 * (diftime / tlength);
    MBperSec = tlength/1024.0/1024.0/diftime;
    printf("# \t %d \t %f \t %ld \t %ld \t %g\n", 
                minor, diftime, phys_size_physmem, tlength, MBperSec);

    tlength = tlengthuser;
  } // minor loop
  exit (EXIT_SUCCESS);
} // main
/*-------------------------------------------------------------*/
/* function for writing to the physmem memory */

void write_physmem(void* addr_user_physmem, long length, int verbose) 
{
  unsigned long *ptr_test;   
  int i;

  ptr_test = (unsigned long *)addr_user_physmem;

  /* Loop over all addresses (sizeof(long) = size of an address on 32 and 64-bit machines) */
  for (i=0; i<length/sizeof(long); i++){
    *(ptr_test+i) = PATTERN + i;
  }
  if (verbose) printf(" + test write: from %#lx, %ld bytes written \n",
                                     (datePointer)addr_user_physmem,length); 
}

/*-------------------------------------------------------------*/
/* function for reading from the physmem memory */

void read_physmem(void* addr_user_physmem, long length, int verbose) 
{
  unsigned long *ptr_test;   
  int i;
  long first_int;
  int flag;

  ptr_test = (unsigned long *)addr_user_physmem;
  first_int = *ptr_test;
  if( first_int != PATTERN ) {
    printf("\n - pattern does not match: found %ld, expected %d \n\n", 
    first_int, PATTERN);
  }
  else
  {
    if (verbose) printf(" + test read:  from %#lx, ", 
                                         (datePointer)addr_user_physmem); 
    flag = 1; // initially ok

    /* Loop over all addresses (sizeof(long) = size of an address on 32 and 64-bit machines) */
    for (i=0; (i<length/sizeof(long)) && (flag == 1); i++) {
    if( *(ptr_test+i) != PATTERN + i )
    flag = 0;
  }
  if (flag == 1 )
  { 
    if (verbose) printf("all %ld bytes are ok \n", length);
  }  
  else 
    printf("\n - not all bytes are ok (%d -> %ld) \n\n", i, *(ptr_test+i) );
  }
}
