/*
 * physmem_lib.h
 *
 * last update: 29-Aug-2012 by Ervin Denes
 */
#ifndef __physmem_lib_h__
#define __physmem_lib_h__

#include "physmem.h"

/* struct to hold info about memory zone offsets*/
typedef struct
{
  unsigned long start;
  unsigned long end;
} offZone_t;

/* struct for physmem handle */
typedef struct
{
  char                   physmem_dev[16];     // physmem device name
  int                    fd;                  // file descriptor
  int                    minor;               // physmem device minor
  int                    full_size;           // 1 if GETFULLSIZE
  int                    num_zones;           // number of memory areas
  unsigned long          phys_addr_physmem;   // start of physmem
  volatile unsigned long *addr_user_physmem;  // user start pointer
  unsigned long          phys_size_physmem;   //
  unsigned long          full_size_physmem;   //
  unsigned long          mapped_size_physmem; //
  unsigned long          sum_gaps;            // total size of memory gaps
  struct memZoneStruct   *zones;              // memory areas
  offZone_t              *phys_offs;          // physical offsets of mem areas
  offZone_t              *user_offs;           // user offsets of memory areas
} physmemHandler_t;

typedef physmemHandler_t* physmemDescriptor;

/* physmem library routines */
int physmemOpen (physmemHandler_t *device, int minor, int full_size, 
                                           unsigned long request_size);

int physmemClose (physmemHandler_t device);

void physmemPrintZones(physmemHandler_t device);

int physmemBlockIsNotContinuous(volatile unsigned int *block_start,
                                volatile unsigned int *block_end,
                                physmemHandler_t device,
                                int                   *start_status,
                                int                   *end_status);

int physmemBlockCross4GB(unsigned long phys_address, long size);

// give physical address offset from user offset
unsigned long physmemPhysOffset(unsigned long user_offset,
                                physmemHandler_t device);

// give user address offset from physical address
unsigned long physmemUserOffset(unsigned long    phys_address,
                                physmemHandler_t device);


#endif

