/* physmem_version.h */

/* driver version number */
#define PHYSMEM_VERSION "6.6"

/*
   versions of kernel module physmem

   v6.6  27-May-2013 Ervin Denes  - add new parameter to physmemOpen: 
                                    requested_size
                                  - add memscan utility program 
                                  - modify driver to map only requestedSize
                                    of physmem memory
                                  - do not load driver if 'nopat' is not
                                    given in boot command  
                                  - modify driver: if there is no "mem" kernel
                                    boot parameter try to find "memmap"
                                  - modifications for kernel > 2.6.36 (Ubuntu)
                                  - modify physmemTest_mod for printing us/Byte
   v6.5  25-Apr-2013 Ervin Denes  - modify install_driver and Makefile, not
                                    using /etc/issue
                                  - Driver gets memory information from BIOS  
                                    e820_saved system instead of /proc/iomem 
                                    in case of SLC6
   v6.4  29-Aug-2012 Ervin Denes  - Change physmem_dev from *char to 
                                    character array in physmemHandler_t type
                                    (physmem_lib.h)
         25-Jul-2012 Gabor Kiss   - Maximum amount of physmem devices set to 12 
                                  - physmemTest_mod - mapping with GETSIZE mode,
                                    write&read into the physmem devices, 
                                    measuring time
                                  - copy install_physmem_driver, 
                                    load_physmem, check_driver and physmem_size
                                    to Linux directory 
         27-Feb-2012 Ervin Denes  - Set minimum linux memory to 500 MB
   v6.3  30-Nov-2011 Ervin Denes  - Do not allow opening more than maxdev device
   v6.2  23-Oct-2011 Ervin Denes  - Delete unnecessary part (memory scanning
                                    for 64 bit machine) of init_module()
   v6.1  12-Oct-2011 Ervin Denes  - Add configuration file
   v5.3  27-Sep-2011 Ervin Denes  - GETFULLSIZE memory mapping bug corrected
                                  - Correct kernel printouts
                                  - Modified physmemTest for full size mapping
   v5.2  25-Aug-2011 Ervin Denes  - Correct checking /proc/iomem area end 
   v5.1  29-Jul-2011 Ervin Denes  - Check if baseaddress falls in 'hole' 
                                  - New library routine physmemBlockCross4GB
         19-Aug-2011              - Expansion to 8 devices
   v5.0  07-Jul-2011 Ervin Denes  - New parameters
                                  - Expansion to more than 2 devices (4 dev's)
   v4.16 11-May-2011  Ervin Denes - Clear allocated memory after kmalloc call
                                                      during RAM test
                                  - Check if /proc/iomem area is passed
                                                           to boot command
                                  - Check if mem parameter is added 
   v4.15 20-Jan-2011  Roberto Divia - Option physmemsize0 can be -1, it means no physmem0 allocated
                      Ervin Denes - Add '-fPIC' compilation flag. It makes possible to link the physmem object library to shared libraries. 
         31-Jan-2011  Roberto Divia - Change printout of total size of phsymem1
   v4.14 25-Nov-2010  Filippo Costa - New printouts of total sizes
   v4.13 05-Mar-2009  Marc Frauman  -  Bug found and being corrected in how the memory is mapped
   v4.12 06-Feb-2009  Marc Frauman  -  Removed support for 2.4 series kernel, first support of >= 2.6.26
   v4.11 05-Feb-2009  Marc Frauman  -  Option to map with or without "virtual" holes corresponding to the physical holes. If ioctl "GETSIZE" is used then only usable memory is mapped, if "GETFULLSIZE" is used, all physical memory is mapped (possibly containing holes)
                                    -  If full mapping is used and the software tries to use the memory in the hole, it will receive a SIGBUS error
   v4.10 16-Dec-2008  Marc Frauman  -  Use /proc/iomem ONLY on 64 bit machines. On 32 bit this file is limited to what Linux can use ...
   v4.9  25-Nov-2008  Marc Frauman  -  Use /proc/iomem to get info about memory zones (necessary to avoid using video memory as RAM memory)
   v4.8  28-Oct-2008  Marc Frauman  -  Removed warnings when compiling on 32-bit PC - Prints correct list of zones
   v4.7  27-Oct-2008  Marc Frauman  -  Modify the ioctl PHYSMEM_GETMEMZONES to give addresses as used by devices and not addresses of physical zones
   v4.6  22-Oct-2008  Marc Frauman  -  Use a compile-time defined list of addresses (beginning of memory zones) - impossible to get one at insert-time (would require to modify the kernel)
   v4.5  06-Oct-2008  Marc Frauman  -  Use dynamic structures as much as possible (no limit on the number of mem zones - we don't know how future machines will work) - still has some limitations (address of memory hole hard-coded)
   v4.4  02-Oct-2008  Marc Frauman  -  Use a structure to get addresses and sizes of the different memory zones
   v4.3  29-Sep-2008  Marc Frauman  -  Merge physmem1 and physmem2 (memory before and after 4GB limit) in one device
   v4.2  29-Sep-2008  Marc Frauman  -  Physical memory over 3.0GB is physically mapped over 4GB
                                    -  Created a 3rd device (minor nr 2) to access it
   v4.2  26-Sep-2008  Marc Frauman  -  remap_page_range and remap_pfn_range don't have same arguments
                                    - works with 64 bits kernel 2.6.26
   v4.1  25-Sep-2008  Marc Frauman  -  Added some includes (with if's) 
                                    - works with 32 bits kernel 2.6.26
   v4.0  24-Sep-2008  Marc Frauman  -  works on SLC4 32bits and 64bits versions
                                    - works on SLC5 32bits version

   V3.4  25-Oct-07  KS     sterring with the module parameters 
   V3.3  19-Jul-07  KS     remove file handling under physmem_mmap
   V3.2  02-Aug-06  KS     modifications to run under SLC3 and SLC4
   V3.1  31-Mar-05  KS     tested with a 2.6 kernel
   V3.0  30-Mar-05  KS     expansion for two devices
                           /dev/physmem0 and /dev/physmem1
   V2.5  01-Jul-03  KS     fix the signature change of remap_page_range
                           for kernel version 2.4.20 onwards
   V2.4  03-Sep-02  KS     use num_physpages instead of high_memory
                           increase physmem maximum to 2 GB
                           improve log messages
   V2.3  21-Aug-02  KS     add parameter physmemsize
                           add schedule() in the scanning loop
                           add check whether ioremap returns NULL
   V2.2  10-Feb-02  KS+AV  customize driver to meet our needs/conventions
   v2.1  01-Feb-02  KS+AV  use driver himemfb as starting point
*/
